#!/usr/bin/perl -w

# ----------------------------------------------------------------------------
# name:         anitmt-client.pl - the anitmt client (NRP protocoll)
# task:         runs an NRP server on the local machine which connects to his
#		server (anitmt-server) for network rendering via NRP 
#               the server ist anitmt-server, current version 0.01
# language:     Perl (developed using perl 5.005_03 on a linux box)
# license:      see file license.eng.txt or webpage 
# maintainer:   Jan Theofel, jan@theofel.de
# version:      c0.01, build 2000-03-12         *initial release*
# new versions: http://www.theofel.de/anitmt/
# ----------------------------------------------------------------------------
# list of hackers:
#   Theofel, Jan (jan@theofel.de)
#     networking, initial system, maintainer, anitmt-server, NRP
# thanks to (alphabetical order):
#   Moser, Manuel (???)
#     maths for the local anitmt rendering system, anitmt demos
#   Trautmann, Martin (martintrautmann@gmx.de)
#     local anitmt rendering system (maintainer & main hacker)
# ----------------------------------------------------------------------------

use IO::Socket;
use Getopt::Long;

# $identify = "anitmt-client  -  an nrp client";
$print_version = 
"anitmt-client  -  a nrp client of the anitmt project\n" .
"  client version:    c0.01\n" .
"  protocoll version: p0.01\n";

&init;
&run;
&done;

sub default_settings
# ----------------------------------------------------------------------------
# task:         sets the default settings, if there are no other settings
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-22
# ----------------------------------------------------------------------------
{
  $cid_file_name	= "anitmt-client.cid"; 
					# where to write/read the CID

  $opt_debug 		= 	1;	# switch debugging on / off ?
  $opt_protdebug	=	1;	# switch debugging on / off ?

  $options{NRP_SERVER} 	= "";		# server isn't a default setting
  $options{NRP_PORT} 	= "4004";	# this is the NRP default port
  $options{SYSTEM} 	= "not defined";# the system we run on is undef
  $options{SPEED} 	= "0 NRPMIPS";	# haven't check it yet
  $options{JOBCACHE} 	= "0 MB";	# no specified limit
  $options{RESULTCACHE} = "0 MB";	# dito
  $options{FTP_SERVER} 	= "";		# we won't use ftp
  $options{FTP_USER} 	= "anitmt";	# defaults ftp user
  $options{FTP_PASSWD} 	= "anitmt";	# and his password
  $options{FTP_SET_BIN} = 1;		# change in bin mode
  $options{FTP_PATH} 	= "";		# this should be user's home dir
  $options{FTP_PATH_IN} = "in-data";	# where do I get data
  $options{FTP_PATH_OUT} = "out-data";	# where do I write data
  $options{NFS_PATH} 	= "";		# we don't use NFS / Samba either
  $options{NFS_PATH_IN} = "in-data";    # where do I get data
  $options{NFS_PATH_OUT} = "out-data";  # where do I write data

  $options{CF_FILE} 	= "./anitmt-client.cf";	# which cf file to read
  $options{TMPPATH}	= "/tmp/anitmt/client/"; # where to store files
  $my_cid 		= "";			# no CID yet

  $options{SLEEP_NO_CONNECT} = 15;	# sleep in sec if server not found
  $options{SLEEP_NO_JOB} = 60;		# sleep in sec when no job found
}

sub read_settings
# ----------------------------------------------------------------------------
# task:         reads in the .cf file and the cid file
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-22
# ----------------------------------------------------------------------------
{
 
  if (-e $cid_file_name)
  {
    open(CIDFILE,"<$cid_file_name") || die "FATAL ERROR: Can't open CID file!";
    $my_cid = <CIDFILE>;
    close(CIDFILE);
  }

  $rr_name = "";
  $last_rr_fp = $last_none;
 
  open(CF,"<$options{CF_FILE}") || die("Fatal error: Can't open cf file!\n");
  while(<CF>)
  {
    $line = $_; 
    chomp($line);
    next if (($line eq "") || ($line =~ /^#/)); 
    $line =~ /^(\S*)\s*=\s*(.*)$/;
    &debug("DEBUG: $1=$2\n"); 
    $setting = $1;
    $value = $2;
    if($setting eq "ADD_RR")
    {
      $rr_name = $value;
      $last_rr_fp = $last_rr;
    }
    elsif($setting eq "ADD_FP")
    {
      $fp_name = $value;
      $last_rr_fp = $last_fp;
    }
    elsif($setting eq "EXEC")
    {
      if($last_rr_fp == $last_rr)
      {
        $rr{$rr_name}->{EXEC} = $value;
      }
      elsif($last_rr_fp == $last_fp)
      {
        $fp{$fp_name}->{ARGS} = $value;
      }
      else
      {
        die "FATAL ERROR: Unmatched EXEC\n";
      }
    }
    elsif($setting eq "ARGS")
    {
      if($last_rr_fp == $last_rr)
      {
        $rr{$rr_name}->{ARGS} = $value;
      }
      elsif($last_rr_fp == $last_fp)
      {
        $fp{$fp_name}->{ARGS} = $value;
      }
      else
      {
        die "FATAL ERROR: Unmatched ARGS\n";
      }
    }
    else
    {
      $options{$setting} = $value;
    }
  }
  close(CF);

  if ($opt_debug)
  {
    foreach $key (keys %rr)
    {
      print "DEBUG: rr ($key)\n";
      print "       exec -> $rr{$key}->{EXEC}\n";
      print "       args -> $rr{$key}->{ARGS}\n";
    }
  }
}

sub parse_commandline
# ----------------------------------------------------------------------------
# task:         reads in the options given in the command line
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  GetOptions(
	"help"			=>	\$opt_help,
	"version"		=>	\$opt_version,

	"debug!"		=>	\$opt_debug,
	"protdebug!"		=>	\$opt_protdebug,
 	
	"kill-cid"		=>	\$opt_kill_cid,
	"renew-cid"		=>	\$opt_renew_cid,

	"cf-file"		=>	\$options{CF_FILE},
	"cid-file"		=>	\$cid_file_name,

	"nrp-server=s"		=>	\$options{NRP_SERVER},
	"nrp-port=i"     	=>      \$options{NRP_PORT},
	"system=s"		=>	\$options{SYSTEM},
	"speed=s"		=>	\$options{SPEED},
	"jobcache=s"		=>	\$options{JOBCACHE},
	"resultcache=s"		=>	\$options{RESULTCACHE},
	"ftp-server=s"		=>	\$options{FTP_SERVER},
	"ftp-user=s"		=>	\$options{FTP_USER},
	"ftp-passwd=s"		=>	\$options{FTP_PASSWD},
	"ftp-set-bin"		=>	\$options{FTP_SET_BIN},
	"ftp-path=s"		=>	\$options{FTP_PATH},
	"ftp-path-in=s"		=>	\$options{FTP_PATH_IN},
	"ftp-path-out=s"	=>	\$options{FTP_PATH_OUT},
	"nfs-path=s"		=>	\$options{NFS_PATH},
	"nfs-path-in=s"		=>	\$options{NFS_PATH_IN},
	"nfs-path-out=s"	=>	\$options{NFS_PATH_OUT},
       
        "sleep-no-job=i"	=>	\$options{SLEEP_NO_JOB},
        "sleep-no-connect=i"	=>	\$options{SLEEP_NO_CONNECT}
  );

  if($opt_help) { &print_usage; }
  if($opt_version) { &print_version; }

  if($opt_kill_cid) { &kill_cid; }
  if($opt_renew_cid) { &renew_cid; }

}

sub kill_cid
# ----------------------------------------------------------------------------
# task:         kills this clients CID from the server and quits
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  die("SORRY, this function is not implemented yet!\n");
}

sub renew_cid
# ----------------------------------------------------------------------------
# task:         kills the old clients CID and resend the informations
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  die("SORRY, this function is not implemented yet!\n");
}

sub print_usage
# ----------------------------------------------------------------------------
# task:         prints out the usage and quits 
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-22
# ----------------------------------------------------------------------------
{

print <<ENDUSAGE;

anitmt-client  -  a nrp client of the anitmt project
visit http://www.theofel.de/anitmt/ for more informations and new versions
written by Jan Theofel (jan\@theofel.de)

Usage:
  --nrp-server          the NRP server you want to connect to (no default)
  --nrp-port		the port you want to use for communication (4004)
  --system		the system the client runs on ("not defined")
  --speed="X NRPMIPS"	defines your speed in NRPMIPS (0 NRPMIPS)
  --jobcache="X MB"	defines your jobcache (0 MB)
  --resultcache="X MB"	defines your resultcache (0 MB)
  --ftp-server		the ftp server for dexm (none)
  --ftp-user		the user to connect to ftp servers (anitmt)
  --ftp-passwd		and his password (anitmt)
  --ftp-set-bin		use binary mode for ftp (yes)
  --ftp-path		path to the data for ftp (none, i.e. users home dir)
  --ftp-path-in		path to incoming data for ftp
  --ftp-path-out	path to outgoing data for ftp
  --nfs-path		path for nfs / samba as dexm (none, i.e. no nfs/samba)
  --nfs-path-in		like above for ftp
  --nfs-path-out	dito

  --help		this short help text
  --version		displays the current version and quits

  --[no]debug		switchs debugging on / off
  --[no]protdebug	switches protocoll debugging on / off

  --kill-cid		kills this client completly from the server
  --nenew-cid		kills the old CID at the server first an then send
			the new configuration of the client for a new CID

  --cf-file		set the cf file (default ./anitmt-client.cf)
  --cid-file		set the cid filename (default ./anitmt-client.cid)
 
  --sleep-no-job        sleep time when server not found (15 sec)
  --sleep-no-connect    sleep time when no job found (60 sec)
ENDUSAGE

die("\n");

}

sub print_version
# ----------------------------------------------------------------------------
# task:         prints out the current version and quits 
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  die($print_version);
}

sub check_settings
# ----------------------------------------------------------------------------
# task:         checks the settings 
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  die("Fatal error: No NRP server given!\n") 
    if ($options{NRP_SERVER} eq "");
  die("Fatal error: No or wrong NRP port number given!\n")
    if ($options{NRP_PORT} eq "")
}

sub init_connection
# ----------------------------------------------------------------------------
# task:         initialises the connection to the server and opens it
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  
  # --- create the communication socket ----
  debug("DEBUG: Connecting to nrp server $options{NRP_SERVER} on port $options{NRP_PORT}\n\n");

  $NRPSOCKET = IO::Socket::INET->new(
    Proto       => "tcp",
    PeerAddr    => $options{NRP_SERVER},
    PeerPort    => $options{NRP_PORT});

  while (!$NRPSOCKET)
  {
    print "DEBUG: No connection. Try again in 15 sec\n";
    sleep $options{SLEEP_NO_CONNET};
    $NRPSOCKET = IO::Socket::INET->new(
      Proto       => "tcp",
      PeerAddr    => $options{NRP_SERVER},
      PeerPort    => $options{NRP_PORT});
  }

  print "DEBUG: Connection established!\n";

  $NRPSOCKET->autoflush(1);

  # --- skip the first server message and send HELO ---
  $continue_data = 1;
  while($continue_data == 1)
  {
    $line = &get_data;
    if(($nrp_status_3 eq "0") || ($nrp_status_3 eq "2"))
    {
      $continue_data = 0;
    }
  }

  &send_data("HELO");
  # --- skip the 2nd server message ---
  $continue_data = 1;
  while($continue_data == 1)
  {
    $line = &get_data;
    if(($nrp_status_3 eq "0") || ($nrp_status_3 eq "2"))
    {
      $continue_data = 0;
    }
  }

}

sub init
# ----------------------------------------------------------------------------
# task:         initialises the complete program 
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-13
# ----------------------------------------------------------------------------
{
  $nrp_status_1 = 0;
  $nrp_status_2 = 0;
  $nrp_status_3 = 0;

  $last_none 	= 0;
  $last_rr	= 1;
  $last_fp	= 2;

  &default_settings;
  &parse_commandline;
  &read_settings;
  &check_settings;

  if(-e $options{TMPPATH})
  {
    system("rm $options{TMPPATH}*");
  }
  else
  {
    system("mkdir $options{TMPPATH}");
  }

  &init_connection;
}

sub send_data
# ----------------------------------------------------------------------------
# task:		sends a line of data to the server
# author:	Jan Theofel, jan@theofel.de
# last changed:	2000-02-06
# ----------------------------------------------------------------------------
{
  my $line_to_send = $_[0];
  exit if ($line_to_send eq "");
  $line_to_send =~ s/[\015\012]+/\015\012/g;
  if ($line_to_send !~ /\015\012$/)
    { $line_to_send = $line_to_send . "\015\012"; }
  print $NRPSOCKET $line_to_send;
  &p_debug($line_to_send);
}

sub get_data
# ----------------------------------------------------------------------------
# task:         gets a line of data from the server
# author:       Jan Theofel, jan@theofel.de
# last changed: 2000-02-06
# ----------------------------------------------------------------------------
{
  $ret_line = <$NRPSOCKET>;
  $ret_line =~ /^(.)(.)(.).*$/;
  $nrp_status_1 = $1;
  $nrp_status_2 = $2;
  $nrp_status_3 = $3;
  &p_debug($ret_line);
  return($ret_line);
}

sub run
# ----------------------------------------------------------------------------
# task:		runs the main part of the program
# author:	Jan Theofel, jan@theofel.de
# last changed:	2000-02-22
# ----------------------------------------------------------------------------
{

  # --- check if this is the first login -> send INFO ---
  if($my_cid eq "")
  {

    $error_count = 0;
    $error = 1;
    while($error_count<=5 && $error) 
    {
      &send_data("INFO");
      $line = &get_data;
      $error = !($nrp_status_1==1 && $nrp_status_2==0 && $nrp_status_3==0);
    }
    if($error_count > 5)
    {
      die("Fatal error: INFO was not accepted by the server!\n");
    }

    &send_data("SYS $options{SYSTEM}");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8)
      { die "Fatal error: SYS not accepted by server!\n"; }

    &send_data("SPED $options{SPEED}");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8)
      { die "Fatal error: SPED not accepted by server!\n"; }

    &send_data("JOBC $options{JOBCACHE}");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8)
      { die "Fatal error: JOBC not accepted by server!\n"; }

    &send_data("RESC $options{RESULTCACHE}");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8)
      { die "Fatal error: RESC not accepted by server!\n"; }

    foreach $key (keys %rr)
    {
      &send_data("ADRR $key");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1)
      { die "Fatal error: ADRR not accepted by server!\n"; }
    }

    &send_data("LARR");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1)
      { die "Fatal error: LARR not accepted by server!\n"; }

    foreach $key (keys %fp)
    {
      &send_data("ADFP $key");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1)
      { die "Fatal error: ADFP not accepted by server!\n"; }
    }

    &send_data("LAFP");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
      { $line = &get_data; }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1)
      { die "Fatal error: LAFP not accepted by server!\n"; }

    if($options{NFS_PATH} ne "")
    {
      &send_data("DEXM NFS");
      $nrp_status_3 = 1;
      while(!($nrp_status_3==0 || $nrp_status_3==2))
        { $line = &get_data; }
      if($nrp_status_2 ne 8)
        { die "Fatal error: DEXM not accepted by server!\n"; }
    }

    &send_data("FIIN");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
    { 
      $line = &get_data; 
      if($line =~ /YOUR CID: \"(.*)\"/)
      {
        $my_cid = $1;
        debug("DEBUG: FOUND MY CID: $my_cid\n");
        open(CIDFILE,">$cid_file_name") || die "FATAL ERROR: Can't open CID file for writing!";
        print CIDFILE $my_cid;
        close(CIDFILE);
      }
    }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
      { die "Fatal error: FIIN not accepted by server!\n"; }

    &send_data("BYE");
    close($NRPSOCKET);

    &init_connection;
  }

  # ------------------------------------ ok, let's get jobs, render them and send back ---
  &send_data("SEHE $my_cid");
  $nrp_status_3 = 1;
  while(!($nrp_status_3==0 || $nrp_status_3==2))
    { $line = &get_data; }
  if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
    { die "Fatal error: SEHE $my_cid not accepted by server!\n"; }

  while(0 == 0)
  {
    &send_data("GET JOB");
    $nrp_status_3 = 1;
    while(!($nrp_status_3==0 || $nrp_status_3==2))
    { 
      $line = &get_data; 
      if($line =~ /^\d\d\d PROJECTNAME:\s*(.*)$/)
      { 
        $job_project_name = $1;
      }
      if($line =~ /^\d\d\d THIS FILE:\s*(.*)$/)
      { 
        $job_this_file = $1;
      }
      if($line =~ /^\d\d\d COPY FILES:\s*(.*)$/)
      { 
        $job_copy_files_string = $1;
        @job_copy_files = split(/,/, $job_copy_files_string);
      }
      if($line =~ /^\d\d\d WIDTH x HEIGHT:\s*(\d*) x (\d*)$/)
      { 
        $job_render_width = $1;
        $job_render_height = $2;
      }
      if($line =~ /^\d\d\d RENDERER:\s*(.*)$/)
      { 
        $job_renderer = $1;
      }
      if($line =~ /^\d\d\d RENDER PARAMS:\s*(.*)$/)
      { 
        $job_render_params = $1;
      }
      if($line =~ /^\d\d\d RENDER FILE:\s*(.*)$/)
      { 
        $job_render_file = $1;
      }
      if($line =~ /^\d\d\d RENDER INC FILE:\s*(.*)$/)
      { 
        $job_render_inc_file = $1;
      }
    }
    if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
    { 
      print "DEBUG: No job found at server, trying agaion later... (60 sec) \n"; 
      &send_data("BYE");
      sleep $options{SLEEP_NO_JOB};
      &init_connection;

     &send_data("SEHE $my_cid");
     $nrp_status_3 = 1;
     while(!($nrp_status_3==0 || $nrp_status_3==2))
       { $line = &get_data; }
     if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
       { die "Fatal error: SEHE $my_cid not accepted by server!\n"; }

    }
    else
    {
      $got_a_job = 1;

      # --- parsing of the settings ---
      $job_render_params = $rr{$job_renderer}->{ARGS} . $job_render_params;
      $job_render_params =~ s/%INFILE%/$job_render_file/;
      $job_out_file = $job_this_file;
      $job_out_file =~ s/\.inc/\.tga/;
      $job_render_params =~ s/%OUTFILE%/$job_out_file/;
      $job_render_params =~ s/%WIDTH%/$job_render_width/;
      $job_render_params =~ s/%HEIGHT%/$job_render_height/;

      print <<DEBUGINFO;
DEBUG job information:
  project name:        $job_project_name
  this file:           $job_this_file
  copy files:          $job_copy_files_string
  render size:         $job_render_width x $job_render_height (w x h)
  renderer:            $job_renderer
  render params:       $job_render_params
  render file:         $job_render_file
  render inc file:     $job_render_inc_file
DEBUGINFO

      foreach $filename (@job_copy_files)
      {
        if(-e "$options{TMPPATH}$filename")
        {
          print "DEBUG: file $options{TMPPATH}$filename found, NO COPY\n";
        }
        else
        {
          print "DEBUG file $options{TMPPATH}$filename not found, getting it\n";
          &send_data("GET FILE $job_project_name $filename");
          $nrp_status_3 = 1;
          while(!($nrp_status_3==0 || $nrp_status_3==2))
            { $line = &get_data; }
          if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
          {
            &send_data("BYE");
            die "FATAL ERROR: Can't get file $filename\n";
          }
          print("DEBUG: cp $options{NFS_PATH_IN}$my_cid/$filename $options{TMPPATH}\n");
          system("cp $options{NFS_PATH_IN}$my_cid/$filename $options{TMPPATH}");
        }
      }

      &send_data("BYE");
      $nrp_status_3 = 1;
      while(!($nrp_status_3==0 || $nrp_status_3==2))
        { $line = &get_data; }
      if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
        { die "Fatal error: BYE not succesful!\n"; }
     }

   # --- creating the include file ---
   open(INC,">$options{TMPPATH}$job_render_inc_file");
   print INC "#include \"$job_this_file\"\n";
   close(INC);

   # --- ok, let's start the renderer ---
   print "DEBUG: Rendering, please wait!\n";
   print("DEBUG: syscall [cd $options{TMPPATH}; $rr{$job_renderer}->{EXEC} $job_render_params -GA$options{TMPPATH}render-out\n");
   system("cd $options{TMPPATH}; $rr{$job_renderer}->{EXEC} $job_render_params -GA$options{TMPPATH}render-out");

   # --- check if the tga output is available ---
   if(-e "$options{TMPPATH}$job_out_file")
   {
     print("DEBUG syscall [cp $options{TMPPATH}$job_out_file $options{NFS_PATH_OUT}$my_cid/$job_out_file]");
     system("cp $options{TMPPATH}$job_out_file $options{NFS_PATH_OUT}$my_cid/$job_out_file");

     &init_connection;

     &send_data("SEHE $my_cid");
     $nrp_status_3 = 1;
     while(!($nrp_status_3==0 || $nrp_status_3==2))
       { $line = &get_data; }
     if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
       { die "Fatal error: SEHE $my_cid not accepted by server!\n"; }

     $nrp_status_2 = 9;
     while($nrp_status_2 eq 9)
     {
       &send_data("SEND $job_project_name $job_out_file");
       $nrp_status_3 = 1;
       while(!($nrp_status_3==0 || $nrp_status_3==2))
         { $line = &get_data; }
       if($nrp_status_2 ne 8 && $nrp_status_2 ne 1 && $nrp_status_2 ne 0)
       { 
         if($nrp_status_2 ne 9)
         {
           die "Fatal error: COULDN'T SEND FILE $job_out_file!\n"; 
         }
       }
     }
   }
   else
   {
     print("FATAL ERROR: Calling renderer seemed to have failed! Check this:\n");
     print "----------------------------------------------------------------\n";
     open(OUT,"<$options{TMPPATH}render-out");
     while(<OUT>)
     { print $_; }
     close(OUT);
     die("\n");
    }
  }
}

sub done
# ----------------------------------------------------------------------------
# task:		quits the program and closes everything correctly
# author:	Jan Theofel, jan@theofel.de
# last changed:	2000-02-22
# ----------------------------------------------------------------------------
{
  close($NRPSOCKET);
  debug("DEBUG: Connection was closed!\n");
}

sub p_debug
# ----------------------------------------------------------------------------
# task:		writes protocoll debug messages
# author:	Jan Theofel, jan@theofel.de
# last changed:	2000-02-22
# ----------------------------------------------------------------------------
{
  if ($opt_protdebug) { print "$_[0]"; } ;
}

sub debug
# ----------------------------------------------------------------------------
# task:		writes protocoll debug messages
# author:	Jan Theofel, jan@theofel.de
# last changed:	2000-02-22
# ----------------------------------------------------------------------------
{
  if ($opt_debug) { print "$_[0]"; } ;
}


